@@ -32,9 +32,8 @@ Revision History:
 #include "Error.h"
 #include "Util.h"
 
-Genome::Genome(GenomeDistance i_maxBases, GenomeDistance nBasesStored, unsigned i_chromosomePadding, unsigned i_maxContigs, bool i_areALTContigsMarked)
-: maxBases(i_maxBases), minLocation(0), maxLocation(i_maxBases), chromosomePadding(i_chromosomePadding), maxContigs(i_maxContigs),
-  mappedFile(NULL)
+Genome::Genome(GenomeDistance i_maxBases, GenomeDistance nBasesStored, unsigned i_chromosomePadding, unsigned i_maxContigs)
+: maxBases(i_maxBases), minLocation(0), maxLocation(i_maxBases), chromosomePadding(i_chromosomePadding), maxContigs(i_maxContigs), mappedFile(NULL)
 {
     bases = ((char *) BigAlloc(nBasesStored + 2 * N_PADDING)) + N_PADDING;
     if (NULL == bases) {
@@ -49,9 +48,9 @@ Genome::Genome(GenomeDistance i_maxBases, GenomeDistance nBasesStored, unsigned
     nBases = 0;
 
     nContigs = 0;
-    contigs = new Contig[maxContigs];
+    contigs = (Contig *)BigAlloc(sizeof(Contig) * maxContigs);
+    contigNumberByOriginalOrder = (InternalContigNum *)BigAlloc(sizeof(InternalContigNum) * maxContigs);
     contigsByName = NULL;
-    areALTContigsMarked = i_areALTContigsMarked;
 }
 
     void
@@ -74,7 +73,7 @@ Genome::addData(const char *data)
 }
 
     void
-Genome::startContig(const char *contigName)
+Genome::startContig(const char *contigName, int originalContigNumber)
 {
     if (nContigs == maxContigs) {
         //
@@ -103,6 +102,7 @@ Genome::startContig(const char *contigName)
     strncpy(contigs[nContigs].name,contigName,len);
     contigs[nContigs].name[len-1] = '\0';
 	contigs[nContigs].isALT = false;
+    contigs[nContigs].originalContigNumber = originalContigNumber;
 
     nContigs++;
 }
@@ -120,14 +120,64 @@ Genome::markContigALT(
 	} // for each contig
 } // markContigALT
 
+    int
+Genome::getContigNumByName(
+    const char* contigName) const
+{
+    for (int i = 0; i < nContigs; i++) {
+        if (!strcmp(contigs[i].name, contigName)) {
+            return i;
+        } // if it matches
+    }
+    return -1;
+}
+
+    GenomeLocation
+Genome::getBeginningLocation(
+    const char* contigName) const
+{
+    int contigNum = getContigNumByName(contigName);
+    return contigs[contigNum].beginningLocation;
+}
+
+    void
+Genome::markContigLiftover(
+    char *contigName,
+    char** altLiftoverContigNames,
+    unsigned* altLiftoverContigFlags,
+    char** altLiftoverProjContigNames,
+    unsigned* altLiftoverProjContigOffsets,
+    char** altLiftoverProjCigar,
+    int nAltLiftover)
+{
+    int contigNum = getContigNumByName(contigName);
+    for (int i = 0; i < nAltLiftover; i++) {
+        if (NULL != altLiftoverContigNames[i]) {
+            if (!strcmp(altLiftoverContigNames[i], contigName)) {
+                int projContigNum = getContigNumByName(altLiftoverProjContigNames[i]);
+                contigs[contigNum].projBeginningLocation = contigs[projContigNum].beginningLocation + altLiftoverProjContigOffsets[i] - 1;
+                contigs[contigNum].isProjRC = (altLiftoverContigFlags[i] & 16) != 0; // bit 4 for rc
+                contigs[contigNum].projCigar = altLiftoverProjCigar[i];
+                return;
+            }
+        }
+    }
+}
+
 Genome::~Genome()
 {
     for (int i = 0; i < nContigs; i++) {
         delete [] contigs[i].name;
         contigs[i].name = NULL;
+        delete [] contigs[i].projCigar;
+        contigs[i].projCigar = NULL;
+        delete [] contigs[i].projCigarOps;
+        contigs[i].projCigarOps = NULL;
     }
 
-    delete [] contigs;
+    BigDealloc(contigs);
+    BigDealloc(contigNumberByOriginalOrder);
+
     if (contigsByName) {
         delete [] contigsByName;
     }
@@ -143,15 +193,17 @@ Genome::~Genome()
 }
 
 // Flags for the options field in the header
-#define	GENOME_FLAG_ALT_CONTIGS_MARKED		0x1
+#define	GENOME_FLAG_ALT_CONTIGS_MARKED		            0x1 // This should always be true now, we dropped support for indices that don't mark ALTs in 1.0.4.  (That doesn't mean that they have to have alts marked, it's just the format.)
 
 // Flags for the per-contig options
 #define GENOME_FLAG_CONTIG_IS_ALT			0x1
+
+#define GENOME_FLAG_ALT_PROJ_CONTIG_IS_RC   0x1
     bool
 Genome::saveToFile(const char *fileName) const
 {
     //
-    // Save file format is (in binary) the number of bases, the number of contigs, followed by
+    // Save file format is the number of bases, the number of contigs and flags followed by
     //  the contigs themselves, rounded up to 4K, followed by the bases.
     //
 
@@ -161,20 +213,18 @@ Genome::saveToFile(const char *fileName) const
         return false;
     } 
 
-    fprintf(saveFile,"%lld %d %d\n",nBases, nContigs, (areALTContigsMarked ? GENOME_FLAG_ALT_CONTIGS_MARKED : 0));	
+    fprintf(saveFile,"%lld %d %d\n",nBases, nContigs, GENOME_FLAG_ALT_CONTIGS_MARKED);	
     char *curChar = NULL;
 
     for (int i = 0; i < nContigs; i++) {
-        for (int n = 0; n < strlen(contigs[i].name); n++){
-         curChar = contigs[i].name + n;
-         if (*curChar == ' '){ *curChar = '_'; }
+        for (int n = 0; n < strlen(contigs[i].name); n++)
+        {
+             curChar = contigs[i].name + n;
+             if (*curChar == ' '){ *curChar = '_'; }
         }
 
-		if (areALTContigsMarked) {
-			fprintf(saveFile, "%lld %x %s\n", GenomeLocationAsInt64(contigs[i].beginningLocation), (contigs[i].isALT ? GENOME_FLAG_CONTIG_IS_ALT : 0), contigs[i].name);
-		} else {
-			fprintf(saveFile, "%lld %s\n", GenomeLocationAsInt64(contigs[i].beginningLocation), contigs[i].name);
-		}
+        fprintf(saveFile, "%lld %x %d %lld %x %d %d %s %s\n", GenomeLocationAsInt64(contigs[i].beginningLocation), (contigs[i].isALT ? GENOME_FLAG_CONTIG_IS_ALT : 0), OriginalContigNumToInt(contigs[i].originalContigNumber), GenomeLocationAsInt64(contigs[i].projBeginningLocation),
+            contigs[i].isProjRC ? GENOME_FLAG_ALT_PROJ_CONTIG_IS_RC : 0, (int) strlen(contigs[i].name), contigs[i].projCigar != NULL ? (int) strlen(contigs[i].projCigar) : 1, contigs[i].name, (contigs[i].projCigar != NULL) ? contigs[i].projCigar : "*");
     }
 
 	//
@@ -202,22 +252,42 @@ Genome::saveToFile(const char *fileName) const
     return true;
 }
 
+    void
+Genome::setUpContigNumbersByOriginalOrder()
+{
+    //
+    // This assumes it's already allocated.
+    //
+
+    memset(contigNumberByOriginalOrder, -1, sizeof(InternalContigNum) * nContigs);
+
+    for (int i = 0; i < nContigs; i++) {
+        OriginalContigNum originalContigNumber = contigs[i].originalContigNumber;
+
+        if (OriginalContigNumToInt(originalContigNumber) < 0 || OriginalContigNumToInt(originalContigNumber) >= nContigs || InternalContigNumToInt(contigNumberByOriginalOrder[OriginalContigNumToInt(originalContigNumber)]) != -1) {
+            WriteErrorMessage("Something's wrong with the original contig numbers in the genome.  If rebuilding the index doesn't help, please submit a bug report\n");
+            soft_exit(1);
+        }
+
+        contigNumberByOriginalOrder[OriginalContigNumToInt(originalContigNumber)] = InternalContigNum(i);
+    } // for each contig
+} // setUpContigNumbersByOriginalOrder
+
     const Genome *
 Genome::loadFromFile(const char *fileName, unsigned chromosomePadding, GenomeLocation minLocation, GenomeDistance length, bool map)
 {    
     GenericFile *loadFile;
     GenomeDistance nBases;
     unsigned nContigs;
-	bool hasALTContigsMarked;
 
-    if (!openFileAndGetSizes(fileName, &loadFile, &nBases, &nContigs, map, &hasALTContigsMarked)) {
+    if (!openFileAndGetSizes(fileName, &loadFile, &nBases, &nContigs, map)) {
         //
         // It already printed an error.  Just fail.
         //
         return NULL;
     }
 
-    GenomeLocation maxLocation(nBases);
+    GenomeLocation maxLocation(nBases); 
 
     if (0 == length) {
         length = maxLocation - minLocation;
@@ -229,24 +299,24 @@ Genome::loadFromFile(const char *fileName, unsigned chromosomePadding, GenomeLoc
         maxLocation = minLocation + length;
     }
 
-    Genome *genome = new Genome(nBases, length, chromosomePadding, nContigs, hasALTContigsMarked);
+    Genome *genome = new Genome(nBases, length, chromosomePadding, nContigs);
    
     genome->nBases = nBases;
     genome->nContigs = genome->maxContigs = nContigs;
-    genome->contigs = new Contig[nContigs];
+    genome->contigs = (Contig *)BigAlloc(sizeof(Contig) * nContigs);
+    genome->contigNumberByOriginalOrder = (InternalContigNum *)BigAlloc(sizeof(InternalContigNum) * nContigs);
+
     genome->minLocation = minLocation;
     if (GenomeLocationAsInt64(minLocation) >= nBases) {
         WriteErrorMessage("Genome::loadFromFile: specified minOffset %u >= nBases %u\n", GenomeLocationAsInt64(minLocation), nBases);
         soft_exit(-1);
     }
-	genome->isContigAware = hasALTContigsMarked;
 
     genome->maxLocation = maxLocation;
 
     int contigNameBufferSize = 0;
     char *contigNameBuffer = NULL;
     size_t n;
-    size_t contigSize;
     char *curName;
     for (unsigned i = 0; i < nContigs; i++) {
         if (NULL == reallocatingFgetsGenericFile(&contigNameBuffer, &contigNameBufferSize, loadFile)) {	 
@@ -256,61 +326,85 @@ Genome::loadFromFile(const char *fileName, unsigned chromosomePadding, GenomeLoc
             return NULL;
         }
 
-        for (n = 0; n < (unsigned)contigNameBufferSize; n++) {
-	        if (contigNameBuffer[n] == ' ') {
-	            contigNameBuffer[n] = '\0'; 
-	            break;
-	        }
-	    }
+        int spacesFound = 0;
+        n = 0;
+        for (;;) {
+            if (contigNameBuffer[n] == ' ') {
+                spacesFound++;
+                if (spacesFound >= 7) {
+                    contigNameBuffer[n] = '\0';
+                    break;
+                } // If we've found all three
+            } // if it's a space
+
+            n++;
+
+            if (n >= contigNameBufferSize) {
+                WriteErrorMessage("Corrupt contig line in genome file found.  Try rebuilding the index.  If that doesn't work, submit a bug report.  Line: '%s'\n", contigNameBuffer);
+                soft_exit(1);
+            }   // if we got to the end without finding enough spaces
+        } // for ever
 
         _int64 contigStart;
-        if (1 != sscanf(contigNameBuffer, "%lld", &contigStart)) {
-            WriteErrorMessage("Unable to parse contig start in genome file '%s', '%s%'\n", fileName, contigNameBuffer);
+        _int64 projContigStart;
+        int contigNameSize;
+        int cigarLength;
+        int originalContigNumber = -1;
+        int flags;
+        int altProjRC;
+
+        if (7 != sscanf(contigNameBuffer, "%lld %x %d %lld %x %d %d", &contigStart, &flags, &originalContigNumber, &projContigStart, &altProjRC, &contigNameSize, &cigarLength)) {
+            WriteErrorMessage("Unable to parse contig start flags or original contig number in genome file '%s', '%s%'\n", fileName, contigNameBuffer);
             soft_exit(1);
         }
 
         genome->contigs[i].beginningLocation = GenomeLocation(contigStart);
+        genome->contigs[i].projBeginningLocation = GenomeLocation(projContigStart);
 	    contigNameBuffer[n] = ' '; 
 	    n++; // increment n so we start copying at the position after the space
 
-		bool contigIsALT = false;
+		genome->contigs[i].isALT = (flags & GENOME_FLAG_CONTIG_IS_ALT) != 0;
+        genome->contigs[i].internalContigNumber = InternalContigNum(i);
+        genome->contigs[i].originalContigNumber = OriginalContigNum(originalContigNumber);
 
-		if (hasALTContigsMarked) {
-			//
-			// There's an extra " %x" after the size with flags.
-			//
-			char *nextSpace = strchr(contigNameBuffer + n, ' ');
-			if (nextSpace == NULL) {
-				WriteErrorMessage("Failure loading genome.  Contigs are supposed to have flags to mark whether they're ALT, but one doesn't.  Line: '%s'\n", contigNameBuffer);
-				soft_exit(1);
-			}
-			*nextSpace = '\0';
-			int flags;
-			if (1 != sscanf(contigNameBuffer + n, "%x", &flags)) {
-				WriteErrorMessage("Failure loading genome.  Unable to parse contig flags in line: '%s'\n", contigNameBuffer);
-				soft_exit(1);
-			}
-
-			if (flags & GENOME_FLAG_CONTIG_IS_ALT) {
-				contigIsALT = true;
-			}
-
-			*nextSpace = ' ';
-			n = nextSpace - contigNameBuffer + 1;
-		}
-
-		genome->contigs[i].isALT = contigIsALT;
-
-	    contigSize = strlen(contigNameBuffer + n) - 1; //don't include the final \n
-        genome->contigs[i].name = new char[contigSize + 1];
-        genome->contigs[i].nameLength = (unsigned)contigSize;
+        genome->contigs[i].name = new char[contigNameSize + 1];
+        genome->contigs[i].nameLength = (unsigned)contigNameSize;
 	    curName = genome->contigs[i].name;
-	    for (unsigned pos = 0; pos < contigSize; pos++) {
+	    for (int pos = 0; pos < contigNameSize; pos++) {
 	      curName[pos] = contigNameBuffer[pos + n];
 	    }
-        curName[contigSize] = '\0';
+        curName[contigNameSize] = '\0';
+
+        n += contigNameSize;
+        contigNameBuffer[n] = ' ';
+        n++;
+
+        if (cigarLength > 1) {
+            genome->contigs[i].projCigar = new char[cigarLength + 1];
+            for (int pos = 0; pos < cigarLength; pos++) {
+                genome->contigs[i].projCigar[pos] = contigNameBuffer[pos + n];
+            }
+            genome->contigs[i].projCigar[cigarLength] = '\0';
+            genome->contigs[i].isProjRC = (altProjRC & GENOME_FLAG_ALT_PROJ_CONTIG_IS_RC) != 0;
+            genome->contigs[i].projCigarOps = new ProjCigarOpFormat[cigarLength];
+            char* nextChunkOfCigar = genome->contigs[i].projCigar;
+            ProjCigarOpFormat* cigarOps = genome->contigs[i].projCigarOps;
+            int scannedOps = 0;
+            while ('\0' != *nextChunkOfCigar) {
+                int fieldsScanned = sscanf(nextChunkOfCigar, "%d%c", &cigarOps[scannedOps].count, &cigarOps[scannedOps].action);
+                scannedOps++;
+                while ('0' <= *nextChunkOfCigar && '9' >= *nextChunkOfCigar) {
+                    nextChunkOfCigar++;
+                }
+                nextChunkOfCigar++;
+                _ASSERT(scannedOps <= cigarLength);
+            }
+            genome->contigs[i].countProjCigarOps = scannedOps;
+        }
     } // for each contig
 
+
+
     if (0 != loadFile->advance(GenomeLocationAsInt64(minLocation))) {
         WriteErrorMessage("Genome::loadFromFile: _fseek64bit failed\n");
         soft_exit(1);
@@ -340,6 +434,7 @@ Genome::loadFromFile(const char *fileName, unsigned chromosomePadding, GenomeLoc
 	
 	genome->fillInContigLengths();
     genome->sortContigsByName();
+    genome->setUpContigNumbersByOriginalOrder();
     delete[] contigNameBuffer;
     return genome;
 }
@@ -382,7 +477,7 @@ Genome::sortContigsByName()
 }
 
     bool
-Genome::openFileAndGetSizes(const char *filename, GenericFile **file, GenomeDistance *nBases, unsigned *nContigs, bool map, bool *hasALTContigsMarked)
+Genome::openFileAndGetSizes(const char *filename, GenericFile **file, GenomeDistance *nBases, unsigned *nContigs, bool map)
 {
 	if (map) {
 		*file = GenericFile_map::open(filename);
@@ -398,7 +493,8 @@ Genome::openFileAndGetSizes(const char *filename, GenericFile **file, GenomeDist
     char linebuf[2000];
     char *retval = (*file)->gets(linebuf, sizeof(linebuf));
 
-    if (NULL == retval || 2 != sscanf(linebuf,"%lld %d\n", nBases, nContigs)) {
+    unsigned flags = 0;
+    if (NULL == retval || 3 != sscanf(linebuf,"%lld %d %d\n", nBases, nContigs, &flags)) {
         (*file)->close();
         delete *file;
         *file = NULL;
@@ -406,34 +502,23 @@ Genome::openFileAndGetSizes(const char *filename, GenericFile **file, GenomeDist
         return false;
     }
 
-	unsigned flags = 0;
-	if (3 == sscanf(linebuf, "%lld %d %d\n", nBases, nContigs, &flags)) 
-	{
-		if (flags & GENOME_FLAG_ALT_CONTIGS_MARKED) 
-		{
-			*hasALTContigsMarked = true;
-		}
-		else 
-		{
-			*hasALTContigsMarked = false;
-		}
-	}
-	else 
-	{
-		*hasALTContigsMarked = false;
-	}
+    if (!(flags & GENOME_FLAG_ALT_CONTIGS_MARKED)) {
+        WriteErrorMessage("Genome doesn't have the ALT contigs marked flag set.  Either it is corrupt or this is a code bug.  Rebuild the index and if this recurs please submit a bug report.\n");
+        soft_exit(1);
+    }
+
     return true;
 }
 
 
     bool 
-Genome::getSizeFromFile(const char *fileName, GenomeDistance *nBases, unsigned *nContigs, bool *hasALTContigsMarked)
+Genome::getSizeFromFile(const char *fileName, GenomeDistance *nBases, unsigned *nContigs)
 {
     GenericFile *file;
     GenomeDistance localNBases;
     unsigned localnContigs;
     
-    if (!openFileAndGetSizes(fileName,&file, nBases ? nBases : &localNBases, nContigs ? nContigs : &localnContigs, false, hasALTContigsMarked)) {
+    if (!openFileAndGetSizes(fileName,&file, nBases ? nBases : &localNBases, nContigs ? nContigs : &localnContigs, false)) {
         return false;
     }
 
@@ -444,7 +529,7 @@ Genome::getSizeFromFile(const char *fileName, GenomeDistance *nBases, unsigned *
 
 
     bool
-Genome::getLocationOfContig(const char *contigName, GenomeLocation *location, int * index) const
+Genome::getLocationOfContig(const char *contigName, GenomeLocation *location, InternalContigNum * index) const
 {
     if (contigsByName) {
         int low = 0;
@@ -457,7 +542,7 @@ Genome::getLocationOfContig(const char *contigName, GenomeLocation *location, in
                     *location = contigsByName[mid].beginningLocation;
                 }
                 if (index != NULL) {
-                    *index = mid;
+                    *index = InternalContigNumToInt(contigsByName[mid].internalContigNumber);
                 }
                 return true;
             } else if (c < 0) {
@@ -468,13 +553,14 @@ Genome::getLocationOfContig(const char *contigName, GenomeLocation *location, in
         }
         return false;
     }
+
     for (int i = 0; i < nContigs; i++) {
         if (!strcmp(contigName,contigs[i].name)) {
             if (NULL != location) {
                 *location = contigs[i].beginningLocation;
             }
 			if (index != NULL) {
-				*index = i;
+				*index = InternalContigNumToInt(i);
 			}
             return true;
         }
@@ -545,6 +631,89 @@ Genome::getNextContigAfterLocation(GenomeLocation location) const
     return NULL; // Should not be reached
 }
 
+int Genome::getProjContigNumAtLocation(GenomeLocation location) const {
+    int contig = getContigNumAtLocation(location);
+    GenomeLocation projLocation = contigs[contig].projBeginningLocation;
+    return getContigNumAtLocation(projLocation);
+}
+
+GenomeDistance Genome::getContigProjSpan(int contig) const {
+    GenomeDistance offset = 0;
+    ProjCigarOpFormat* cigar = contigs[contig].projCigarOps;
+    int nCigarOps = contigs[contig].countProjCigarOps;
+    if (nCigarOps > 0) {
+        char action = cigar[0].action;
+        int count = cigar[0].count;
+        if (action != 'S' && action != 'H') {
+            offset += count;
+        }
+    }
+    for (int i = 1; i < nCigarOps; ++i) {
+        char action = cigar[i].action;
+        int count = cigar[i].count;
+        if (action == 'M' || action == 'I') {
+            offset += count;
+        }
+    }
+    return offset;
+}
+
+GenomeLocation Genome::getProjLocation(GenomeLocation location, int alnSpan) const {
+    int contig = getContigNumAtLocation(location);
+    bool isProjRC = contigs[contig].isProjRC;
+    GenomeDistance offset = !isProjRC ? location - contigs[contig].beginningLocation : getContigProjSpan(contig) - (location - contigs[contig].beginningLocation + alnSpan);
+    GenomeDistance projOffset = 0;
+    ProjCigarOpFormat* cigar = contigs[contig].projCigarOps;
+    int cigarOpCount = contigs[contig].countProjCigarOps;
+    int start = 0;
+    if (cigarOpCount > 0) {
+        char action = cigar[0].action;
+        if (action == 'S' || action == 'H') {
+            start = 1;
+        }
+    }
+    // From bwa-postalt.js: https://github.com/lh3/bwa/blob/master/bwakit/bwa-postalt.js
+    GenomeDistance x = 0, y = 0; // x - REF, y - ALT
+    for (int i = start; i < cigarOpCount; ++i) {
+        char action = cigar[i].action;
+        int count = cigar[i].count;
+        if (action == 'M') {
+            if (y <= offset && offset < y + count) {
+                projOffset = x + (offset - y);
+                break;
+            }
+            x += count;
+            y += count;
+        }
+        else if (action == 'D') {
+            x += count;
+        }
+        else if (action == 'I') {
+            if (y <= offset && offset < y + count) {
+                projOffset = x;
+                break;
+        }
+            y += count;
+        }
+        else if (action == 'S' || action == 'H') {
+            if (y <= offset && offset < y + count) {
+                projOffset = -1;
+                break;
+            }
+            y += count;
+        }
+    }
+    if (projOffset == -1) {
+        return InvalidGenomeLocation;
+    }
+    return contigs[contig].projBeginningLocation + projOffset;
+}
+
+bool Genome::isProjContigRC(GenomeLocation location) const {
+    int contig = getContigNumAtLocation(location);
+    return contigs[contig].isProjRC;
+}
+
 GenomeDistance DistanceBetweenGenomeLocations(GenomeLocation locationA, GenomeLocation locationB) 
 {
     if (locationA > locationB) return locationA - locationB;
